@@ -1,6 +1,6 @@
 /*
  * BlueALSA - main.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -13,6 +13,7 @@
 #endif
 
 #include <getopt.h>
+#include <sched.h>
 #include <signal.h>
 #include <stdbool.h>
 #include <stdint.h>
@@ -20,6 +21,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
+#include <syslog.h>
 #include <time.h>
 
 #include <gio/gio.h>
@@ -33,8 +35,7 @@
 #include "a2dp.h"
 #include "a2dp-sbc.h"
 #include "audio.h"
-#include "ba-adapter.h"
-#include "bluealsa-config.h"
+#include "ba-config.h"
 #include "bluealsa-dbus.h"
 #include "bluealsa-iface.h"
 #include "bluez.h"
@@ -94,6 +95,9 @@ static char *get_hfp_codecs(void) {
 		hfp_codec_id_to_string(HFP_CODEC_CVSD),
 #if ENABLE_MSBC
 		hfp_codec_id_to_string(HFP_CODEC_MSBC),
+#endif
+#if ENABLE_LC3_SWB
+		hfp_codec_id_to_string(HFP_CODEC_LC3_SWB),
 #endif
 		NULL,
 	};
@@ -150,12 +154,15 @@ int main(int argc, char **argv) {
 		{ "help", no_argument, NULL, 'h' },
 		{ "version", no_argument, NULL, 'V' },
 		{ "syslog", no_argument, NULL, 'S' },
+		{ "loglevel", required_argument, NULL, 23 },
 		{ "dbus", required_argument, NULL, 'B' },
 		{ "device", required_argument, NULL, 'i' },
 		{ "profile", required_argument, NULL, 'p' },
 		{ "codec", required_argument, NULL, 'c' },
 		{ "initial-volume", required_argument, NULL, 17 },
 		{ "keep-alive", required_argument, NULL, 8 },
+		{ "io-rt-priority", required_argument, NULL, 3 },
+		{ "disable-realtek-usb-fix", no_argument, NULL, 21 },
 		{ "a2dp-force-mono", no_argument, NULL, 6 },
 		{ "a2dp-force-audio-cd", no_argument, NULL, 7 },
 		{ "a2dp-volume", no_argument, NULL, 9 },
@@ -177,6 +184,9 @@ int main(int argc, char **argv) {
 #if ENABLE_MP3LAME
 		{ "mp3-algorithm", required_argument, NULL, 12 },
 		{ "mp3-vbr-quality", required_argument, NULL, 13 },
+#endif
+#if ENABLE_MIDI
+		{ "midi-advertisement", no_argument, NULL, 22 },
 #endif
 		{ "xapl-resp-name", required_argument, NULL, 16 },
 		{ 0, 0, 0, 0 },
@@ -198,12 +208,15 @@ int main(int argc, char **argv) {
 					"  -h, --help\t\t\tprint this help and exit\n"
 					"  -V, --version\t\t\tprint version and exit\n"
 					"  -S, --syslog\t\t\tsend output to syslog\n"
+					"  --loglevel=LEVEL\t\tminimum message priority\n"
 					"  -B, --dbus=NAME\t\tD-Bus service name suffix\n"
 					"  -i, --device=hciX\t\tHCI device(s) to use\n"
 					"  -p, --profile=NAME\t\tset enabled BT profiles\n"
 					"  -c, --codec=NAME\t\tset enabled BT audio codecs\n"
 					"  --initial-volume=NUM\t\tinitial volume level [0-100]\n"
 					"  --keep-alive=SEC\t\tkeep Bluetooth transport alive\n"
+					"  --io-rt-priority=NUM\t\treal-time priority for IO threads\n"
+					"  --disable-realtek-usb-fix\tdisable fix for mSBC on Realtek USB\n"
 					"  --a2dp-force-mono\t\ttry to force monophonic sound\n"
 					"  --a2dp-force-audio-cd\t\ttry to force 44.1 kHz sampling\n"
 					"  --a2dp-volume\t\t\tnative volume control by default\n"
@@ -225,18 +238,24 @@ int main(int argc, char **argv) {
 #if ENABLE_MP3LAME
 					"  --mp3-algorithm=TYPE\t\tselect LAME encoder algorithm type\n"
 					"  --mp3-vbr-quality=MODE\tset LAME encoder VBR quality mode\n"
+#endif
+#if ENABLE_MIDI
+					"  --midi-advertisement\t\tenable LE advertisement for BLE-MIDI\n"
 #endif
 					"  --xapl-resp-name=NAME\t\tset product name used by XAPL\n"
 					"\nAvailable BT profiles:\n"
-					"  - a2dp-source\tAdvanced Audio Source (%s)\n"
-					"  - a2dp-sink\tAdvanced Audio Sink (%s)\n"
+					"  - a2dp-source\tAdvanced Audio Source (v1.4)\n"
+					"  - a2dp-sink\tAdvanced Audio Sink (v1.4)\n"
 #if ENABLE_OFONO
 					"  - hfp-ofono\tHands-Free AG/HF handled by oFono\n"
 #endif
-					"  - hfp-ag\tHands-Free Audio Gateway (%s)\n"
-					"  - hfp-hf\tHands-Free (%s)\n"
-					"  - hsp-ag\tHeadset Audio Gateway (%s)\n"
-					"  - hsp-hs\tHeadset (%s)\n"
+					"  - hfp-ag\tHands-Free Audio Gateway (v1.9)\n"
+					"  - hfp-hf\tHands-Free (v1.9)\n"
+					"  - hsp-ag\tHeadset Audio Gateway (v1.2)\n"
+					"  - hsp-hs\tHeadset (v1.2)\n"
+#if ENABLE_MIDI
+					"  - midi\tBluetooth LE MIDI (v1.0)\n"
+#endif
 					"\n"
 					"Available BT audio codecs:\n"
 					"  a2dp-source:\t%s\n"
@@ -244,9 +263,6 @@ int main(int argc, char **argv) {
 					"  hfp-*:\t%s\n"
 					"",
 					argv[0],
-					"v1.3", "v1.3",
-					"v1.7", "v1.7",
-					"v1.2", "v1.2",
 					get_a2dp_codecs(A2DP_SOURCE),
 					get_a2dp_codecs(A2DP_SINK),
 					get_hfp_codecs());
@@ -263,7 +279,7 @@ int main(int argc, char **argv) {
 
 	log_open(basename(argv[0]), syslog);
 
-	if (bluealsa_config_init() != 0) {
+	if (ba_config_init() != 0) {
 		error("Couldn't initialize bluealsa config");
 		return EXIT_FAILURE;
 	}
@@ -277,6 +293,28 @@ int main(int argc, char **argv) {
 		case 'S' /* --syslog */ :
 			break;
 
+		case 23 /* --loglevel=LEVEL */ : {
+
+			static const nv_entry_t values[] = {
+				{ "error", .v.ui = LOG_ERR },
+				{ "warning", .v.ui = LOG_WARNING },
+				{ "info", .v.ui = LOG_INFO },
+#if DEBUG
+				{ "debug", .v.ui = LOG_DEBUG },
+#endif
+				{ 0 },
+			};
+
+			const nv_entry_t *entry;
+			if ((entry = nv_find(values, optarg)) == NULL) {
+				error("Invalid loglevel {%s}: %s", nv_join_names(values), optarg);
+				return EXIT_FAILURE;
+			}
+
+			log_set_min_priority(entry->v.ui);
+			break;
+		}
+
 		case 'B' /* --dbus=NAME */ :
 			snprintf(dbus_service, sizeof(dbus_service), BLUEALSA_SERVICE ".%s", optarg);
 			if (!g_dbus_is_name(dbus_service)) {
@@ -304,6 +342,9 @@ int main(int argc, char **argv) {
 				{ "hfp-ag", &config.profile.hfp_ag },
 				{ "hsp-hs", &config.profile.hsp_hs },
 				{ "hsp-ag", &config.profile.hsp_ag },
+#if ENABLE_MIDI
+				{ "midi", &config.profile.midi },
+#endif
 			};
 
 			bool matched = false;
@@ -325,12 +366,15 @@ int main(int argc, char **argv) {
 		case 'c' /* --codec=NAME */ : {
 
 			static const struct {
-				uint16_t codec_id;
+				uint32_t codec_id;
 				bool *ptr;
 			} hfp_codecs[] = {
 				{ HFP_CODEC_CVSD, &config.hfp.codecs.cvsd },
 #if ENABLE_MSBC
 				{ HFP_CODEC_MSBC, &config.hfp.codecs.msbc },
+#endif
+#if ENABLE_LC3_SWB
+				{ HFP_CODEC_LC3_SWB, &config.hfp.codecs.lc3_swb },
 #endif
 			};
 
@@ -342,7 +386,7 @@ int main(int argc, char **argv) {
 			}
 
 			struct a2dp_codec * const * cc = a2dp_codecs;
-			uint16_t codec_id = a2dp_codecs_codec_id_from_string(optarg);
+			uint32_t codec_id = a2dp_codecs_codec_id_from_string(optarg);
 			for (struct a2dp_codec *c = *cc; c != NULL; c = *++cc)
 				if (c->codec_id == codec_id) {
 					c->enabled = enable;
@@ -379,6 +423,20 @@ int main(int argc, char **argv) {
 			config.keep_alive_time = atof(optarg) * 1000;
 			break;
 
+		case 3 /* --io-rt-priority=NUM */ :
+			config.io_thread_rt_priority = atoi(optarg);
+			const int min = sched_get_priority_min(SCHED_FIFO);
+			const int max = sched_get_priority_max(SCHED_FIFO);
+			if (config.io_thread_rt_priority < min || max < config.io_thread_rt_priority) {
+				error("Invalid IO thread RT priority [%d, %d]: %s", min, max, optarg);
+				return EXIT_FAILURE;
+			}
+			break;
+
+		case 21 /* --disable-realtek-usb-fix */ :
+			config.disable_realtek_usb_fix = true;
+			break;
+
 		case 6 /* --a2dp-force-mono */ :
 			config.a2dp.force_mono = true;
 			break;
@@ -511,6 +569,12 @@ int main(int argc, char **argv) {
 		}
 #endif
 
+#if ENABLE_MIDI
+		case 22 /* --midi-advertisement */ :
+			config.midi.advertise = true;
+			break;
+#endif
+
 		case 16 /* --xapl-resp-name=NAME */ :
 			config.hfp.xapl_product_name = optarg;
 			break;
@@ -524,7 +588,7 @@ int main(int argc, char **argv) {
 	if (!(config.profile.a2dp_source || config.profile.a2dp_sink ||
 				config.profile.hfp_hf || config.profile.hfp_ag ||
 				config.profile.hsp_hs || config.profile.hsp_ag ||
-				config.profile.hfp_ofono)) {
+				config.profile.hfp_ofono || config.profile.midi)) {
 		error("It is required to enabled at least one BT profile");
 		fprintf(stderr, "Try '%s --help' for more information.\n", argv[0]);
 		return EXIT_FAILURE;
@@ -569,7 +633,8 @@ int main(int argc, char **argv) {
 	a2dp_sbc_sink.enabled = true;
 	config.hfp.codecs.cvsd = true;
 
-	a2dp_codecs_init();
+	if (a2dp_codecs_init() == -1)
+		return EXIT_FAILURE;
 
 	const char *storage_base_dir = BLUEALSA_STORAGE_DIR;
 #if ENABLE_SYSTEMD
@@ -599,8 +664,7 @@ int main(int argc, char **argv) {
 	g_main_loop_run(loop);
 
 	/* cleanup internal structures */
-	for (size_t i = 0; i < ARRAYSIZE(config.adapters); i++)
-		ba_adapter_destroy(config.adapters[i]);
+	bluez_destroy();
 
 	storage_destroy();
 	g_dbus_connection_close_sync(config.dbus, NULL, NULL);
