@@ -1,6 +1,6 @@
 /*
  * BlueALSA - sco.h
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -8,6 +8,7 @@
  *
  */
 
+#pragma once
 #ifndef BLUEALSA_SCO_H_
 #define BLUEALSA_SCO_H_
 
@@ -19,6 +20,7 @@
 #include "ba-transport.h"
 
 int sco_setup_connection_dispatcher(struct ba_adapter *a);
-void *sco_thread(struct ba_transport_thread *th);
+void sco_transport_init(struct ba_transport *t);
+int sco_transport_start(struct ba_transport *t);
 
 #endif
