@@ -1,6 +1,6 @@
 /*
  * BlueALSA - aplay.c
- * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2016-2024 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -24,6 +24,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
+#include <sys/eventfd.h>
 #include <sys/param.h>
 #include <unistd.h>
 
@@ -32,13 +34,22 @@
 #include <dbus/dbus.h>
 
 #include "shared/dbus-client.h"
+#include "shared/dbus-client-pcm.h"
 #include "shared/defs.h"
 #include "shared/ffb.h"
 #include "shared/log.h"
+#include "shared/nv.h"
 #include "alsa-mixer.h"
 #include "alsa-pcm.h"
 #include "dbus.h"
 
+enum volume_type {
+	VOL_TYPE_AUTO,
+	VOL_TYPE_MIXER,
+	VOL_TYPE_SOFTWARE,
+	VOL_TYPE_NONE,
+};
+
 struct io_worker {
 	pthread_t thread;
 	/* used BlueALSA PCM device */
@@ -63,6 +74,7 @@ static unsigned int verbose = 0;
 static bool list_bt_devices = false;
 static bool list_bt_pcms = false;
 static const char *pcm_device = "default";
+static enum volume_type volume_type = VOL_TYPE_AUTO;
 static const char *mixer_device = "default";
 static const char *mixer_elem_name = "Master";
 static unsigned int mixer_elem_index = 0;
@@ -90,15 +102,9 @@ static struct io_worker *workers = NULL;
 static size_t workers_count = 0;
 static size_t workers_size = 0;
 
-static atomic_bool main_loop_on = true;
+static int main_loop_quit_event_fd = -1;
 static void main_loop_stop(int sig) {
-	/* Call to this handler restores the default action, so on the
-	 * second call the program will be forcefully terminated. */
-
-	struct sigaction sigact = { .sa_handler = SIG_DFL };
-	sigaction(sig, &sigact, NULL);
-
-	main_loop_on = false;
+	eventfd_write(main_loop_quit_event_fd, sig);
 }
 
 static int parse_bt_addresses(char *argv[], size_t count) {
@@ -246,17 +252,32 @@ static void print_bt_pcm_list(void) {
 
 }
 
-static struct ba_pcm *get_ba_pcm(const char *path) {
+static struct ba_pcm *ba_pcm_add(const struct ba_pcm *pcm) {
+	struct ba_pcm *tmp;
+	if ((tmp = realloc(ba_pcms, (ba_pcms_count + 1) * sizeof(*ba_pcms))) == NULL)
+		return NULL;
+	ba_pcms = tmp;
+	memcpy(&ba_pcms[ba_pcms_count], pcm, sizeof(*ba_pcms));
+	return &ba_pcms[ba_pcms_count++];
+}
 
-	size_t i;
-
-	for (i = 0; i < ba_pcms_count; i++)
+static struct ba_pcm *ba_pcm_get(const char *path) {
+	for (size_t i = 0; i < ba_pcms_count; i++)
 		if (strcmp(ba_pcms[i].pcm_path, path) == 0)
 			return &ba_pcms[i];
-
 	return NULL;
 }
 
+static void ba_pcm_remove(const char *path) {
+	for (size_t i = 0; i < ba_pcms_count; i++)
+		if (strcmp(ba_pcms[i].pcm_path, path) == 0) {
+			memmove(&ba_pcms[i], &ba_pcms[i + 1],
+					(ba_pcms_count - i - 1) * sizeof(*ba_pcms));
+			ba_pcms_count--;
+			break;
+		}
+}
+
 static struct io_worker *get_active_io_worker(void) {
 
 	struct io_worker *w = NULL;
@@ -362,7 +383,7 @@ static int io_worker_mixer_volume_sync_ba_pcm(
 	ba_pcm->volume.ch2_volume = volume;
 
 	DBusError err = DBUS_ERROR_INIT;
-	if (!bluealsa_dbus_pcm_update(&dbus_ctx, ba_pcm, BLUEALSA_PCM_VOLUME, &err)) {
+	if (!ba_dbus_pcm_update(&dbus_ctx, ba_pcm, BLUEALSA_PCM_VOLUME, &err)) {
 		error("Couldn't update BlueALSA source PCM: %s", err.message);
 		dbus_error_free(&err);
 		return -1;
@@ -496,8 +517,25 @@ static void *io_worker_routine(struct io_worker *w) {
 	}
 
 	DBusError err = DBUS_ERROR_INIT;
+
+	/* initialize the PCM soft_volume setting */
+	if (volume_type != VOL_TYPE_AUTO) {
+		bool softvol = (volume_type == VOL_TYPE_SOFTWARE);
+		debug("Setting BlueALSA source PCM volume mode: %s: %s",
+				w->ba_pcm.pcm_path, softvol ? "software" : "pass-through");
+		if (softvol != w->ba_pcm.soft_volume) {
+			w->ba_pcm.soft_volume = softvol;
+			if (!ba_dbus_pcm_update(&dbus_ctx, &w->ba_pcm,
+						BLUEALSA_PCM_SOFT_VOLUME, &err)) {
+				error("Couldn't set BlueALSA source PCM volume mode: %s", err.message);
+				dbus_error_free(&err);
+				goto fail;
+			}
+		}
+	}
+
 	debug("Opening BlueALSA source PCM: %s", w->ba_pcm.pcm_path);
-	if (!bluealsa_dbus_pcm_open(&dbus_ctx, w->ba_pcm.pcm_path,
+	if (!ba_dbus_pcm_open(&dbus_ctx, w->ba_pcm.pcm_path,
 				&w->ba_pcm_fd, &w->ba_pcm_ctrl_fd, &err)) {
 		error("Couldn't open BlueALSA source PCM: %s", err.message);
 		dbus_error_free(&err);
@@ -523,15 +561,17 @@ static void *io_worker_routine(struct io_worker *w) {
 	int timeout = -1;
 
 	debug("Starting IO loop");
-	while (main_loop_on) {
+	for (;;) {
 
 		if (single_playback_mutex_locked) {
 			pthread_mutex_unlock(&single_playback_mutex);
 			single_playback_mutex_locked = false;
 		}
 
-		struct pollfd fds[16] = {{ w->ba_pcm_fd, POLLIN, 0 }};
-		nfds_t nfds = 1;
+		struct pollfd fds[16] = {
+			{ main_loop_quit_event_fd, POLLIN, 0 },
+			{ w->ba_pcm_fd, POLLIN, 0 }};
+		nfds_t nfds = 2;
 
 		if (w->snd_mixer != NULL)
 			nfds += snd_mixer_poll_descriptors_count(w->snd_mixer);
@@ -542,7 +582,7 @@ static void *io_worker_routine(struct io_worker *w) {
 		}
 
 		if (w->snd_mixer != NULL)
-			snd_mixer_poll_descriptors(w->snd_mixer, fds + 1, nfds - 1);
+			snd_mixer_poll_descriptors(w->snd_mixer, fds + 2, nfds - 2);
 
 		/* Reading from the FIFO won't block unless there is an open connection
 		 * on the writing side. However, the server does not open PCM FIFO until
@@ -568,11 +608,14 @@ static void *io_worker_routine(struct io_worker *w) {
 			goto close_alsa;
 		}
 
+		if (fds[0].revents & POLLIN)
+			break;
+
 		if (w->snd_mixer != NULL)
 			snd_mixer_handle_events(w->snd_mixer);
 
 		size_t read_samples = 0;
-		if (fds[0].revents & POLLIN) {
+		if (fds[1].revents & POLLIN) {
 
 			ssize_t ret;
 			size_t _in = MIN(pcm_max_read_len, ffb_blen_in(&buffer));
@@ -588,13 +631,13 @@ static void *io_worker_routine(struct io_worker *w) {
 				warn("Invalid read from BlueALSA source PCM: %zd %% %zd != 0", ret, pcm_format_size);
 
 		}
-		else if (fds[0].revents & POLLHUP) {
+		else if (fds[1].revents & POLLHUP) {
 			/* source PCM FIFO has been terminated on the writing side */
 			debug("BlueALSA source PCM disconnected: %s", w->ba_pcm.pcm_path);
 			break;
 		}
-		else if (fds[0].revents)
-			error("Unexpected BlueALSA source PCM poll event: %#x", fds[0].revents);
+		else if (fds[1].revents)
+			error("Unexpected BlueALSA source PCM poll event: %#x", fds[1].revents);
 
 		if (read_samples == 0)
 			continue;
@@ -661,12 +704,14 @@ static void *io_worker_routine(struct io_worker *w) {
 			snd_pcm_get_params(w->snd_pcm, &buffer_frames, &period_frames);
 			pcm_max_read_len = period_frames * w->ba_pcm.channels * pcm_format_size;
 
-			debug("Opening ALSA mixer: name=%s elem=%s index=%u",
-					mixer_device, mixer_elem_name, mixer_elem_index);
-			if (alsa_mixer_open(&w->snd_mixer, &w->snd_mixer_elem,
-						mixer_device, mixer_elem_name, mixer_elem_index, &tmp) != 0) {
-				warn("Couldn't open ALSA mixer: %s", tmp);
-				free(tmp);
+			if (mixer_device != NULL) {
+				debug("Opening ALSA mixer: name=%s elem=%s index=%u",
+						mixer_device, mixer_elem_name, mixer_elem_index);
+				if (alsa_mixer_open(&w->snd_mixer, &w->snd_mixer_elem,
+							mixer_device, mixer_elem_name, mixer_elem_index, &tmp) != 0) {
+					warn("Couldn't open ALSA mixer: %s", tmp);
+					free(tmp);
+				}
 			}
 
 			io_worker_mixer_volume_sync_setup(w);
@@ -904,20 +949,17 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 				goto fail;
 			struct ba_pcm pcm;
 			DBusError err = DBUS_ERROR_INIT;
-			if (!bluealsa_dbus_message_iter_get_pcm(&iter, &err, &pcm)) {
+			if (!dbus_message_iter_get_ba_pcm(&iter, &err, &pcm)) {
 				error("Couldn't add new BlueALSA PCM: %s", err.message);
 				dbus_error_free(&err);
 				goto fail;
 			}
 			if (pcm.transport == BA_PCM_TRANSPORT_NONE)
 				goto fail;
-			struct ba_pcm *tmp = ba_pcms;
-			if ((ba_pcms = realloc(ba_pcms, (ba_pcms_count + 1) * sizeof(*ba_pcms))) == NULL) {
-				error("Couldn't add new BlueALSA PCM: %s", strerror(ENOMEM));
-				ba_pcms = tmp;
+			if (ba_pcm_add(&pcm) == NULL) {
+				error("Couldn't add new BlueALSA PCM: %s", strerror(errno));
 				goto fail;
 			}
-			memcpy(&ba_pcms[ba_pcms_count++], &pcm, sizeof(*ba_pcms));
 			supervise_io_worker(&pcm);
 			return DBUS_HANDLER_RESULT_HANDLED;
 		}
@@ -930,9 +972,10 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 			}
 			dbus_message_iter_get_basic(&iter, &path);
 			struct ba_pcm *pcm;
-			if ((pcm = get_ba_pcm(path)) == NULL)
+			if ((pcm = ba_pcm_get(path)) == NULL)
 				goto fail;
 			supervise_io_worker_stop(pcm);
+			ba_pcm_remove(path);
 			return DBUS_HANDLER_RESULT_HANDLED;
 		}
 
@@ -940,7 +983,7 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 
 	if (strcmp(interface, DBUS_INTERFACE_PROPERTIES) == 0) {
 		struct ba_pcm *pcm;
-		if ((pcm = get_ba_pcm(path)) == NULL)
+		if ((pcm = ba_pcm_get(path)) == NULL)
 			goto fail;
 		if (!dbus_message_iter_init(message, &iter) ||
 				dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING) {
@@ -949,7 +992,7 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 		}
 		dbus_message_iter_get_basic(&iter, &interface);
 		dbus_message_iter_next(&iter);
-		if (!bluealsa_dbus_message_iter_get_pcm_props(&iter, NULL, pcm))
+		if (!dbus_message_iter_get_ba_pcm_props(&iter, NULL, pcm))
 			goto fail;
 		if ((worker = supervise_io_worker(pcm)) != NULL)
 			io_worker_mixer_volume_sync_snd_mixer_elem(worker, pcm);
@@ -968,6 +1011,7 @@ int main(int argc, char *argv[]) {
 		{ "help", no_argument, NULL, 'h' },
 		{ "version", no_argument, NULL, 'V' },
 		{ "syslog", no_argument, NULL, 'S' },
+		{ "loglevel", required_argument, NULL, 9 },
 		{ "verbose", no_argument, NULL, 'v' },
 		{ "list-devices", no_argument, NULL, 'l' },
 		{ "list-pcms", no_argument, NULL, 'L' },
@@ -975,6 +1019,7 @@ int main(int argc, char *argv[]) {
 		{ "pcm", required_argument, NULL, 'D' },
 		{ "pcm-buffer-time", required_argument, NULL, 3 },
 		{ "pcm-period-time", required_argument, NULL, 4 },
+		{ "volume", required_argument, NULL, '8' },
 		{ "mixer-device", required_argument, NULL, 'M' },
 		{ "mixer-name", required_argument, NULL, 6 },
 		{ "mixer-index", required_argument, NULL, 7 },
@@ -985,6 +1030,7 @@ int main(int argc, char *argv[]) {
 	};
 
 	bool syslog = false;
+	const char *volume_type_str = "auto";
 
 	/* Check if syslog forwarding has been enabled. This check has to be
 	 * done before anything else, so we can log early stage warnings and
@@ -999,6 +1045,7 @@ int main(int argc, char *argv[]) {
 					"  -h, --help\t\t\tprint this help and exit\n"
 					"  -V, --version\t\t\tprint version and exit\n"
 					"  -S, --syslog\t\t\tsend output to syslog\n"
+					"  --loglevel=LEVEL\t\tminimum message priority\n"
 					"  -v, --verbose\t\t\tmake output more verbose\n"
 					"  -l, --list-devices\t\tlist available BT audio devices\n"
 					"  -L, --list-pcms\t\tlist available BT audio PCMs\n"
@@ -1006,9 +1053,10 @@ int main(int argc, char *argv[]) {
 					"  -D, --pcm=NAME\t\tplayback PCM device to use\n"
 					"  --pcm-buffer-time=INT\t\tplayback PCM buffer time\n"
 					"  --pcm-period-time=INT\t\tplayback PCM period time\n"
+					"  --volume=TYPE\t\t\tvolume control type [auto|mixer|none|software]\n"
 					"  -M, --mixer-device=NAME\tmixer device to use\n"
 					"  --mixer-name=NAME\t\tmixer element name\n"
-					"  --mixer-index=NUM\t\tmixer element channel index\n"
+					"  --mixer-index=NUM\t\tmixer element index\n"
 					"  --profile-a2dp\t\tuse A2DP profile (default)\n"
 					"  --profile-sco\t\t\tuse SCO profile\n"
 					"  --single-audio\t\tsingle audio mode\n"
@@ -1046,6 +1094,28 @@ int main(int argc, char *argv[]) {
 		case 'v' /* --verbose */ :
 			break;
 
+		case 9 /* --loglevel=LEVEL */ : {
+
+			static const nv_entry_t values[] = {
+				{ "error", .v.ui = LOG_ERR },
+				{ "warning", .v.ui = LOG_WARNING },
+				{ "info", .v.ui = LOG_INFO },
+#if DEBUG
+				{ "debug", .v.ui = LOG_DEBUG },
+#endif
+				{ 0 },
+			};
+
+			const nv_entry_t *entry;
+			if ((entry = nv_find(values, optarg)) == NULL) {
+				error("Invalid loglevel {%s}: %s", nv_join_names(values), optarg);
+				return EXIT_FAILURE;
+			}
+
+			log_set_min_priority(entry->v.ui);
+			break;
+		}
+
 		case 'l' /* --list-devices */ :
 			list_bt_devices = true;
 			break;
@@ -1071,6 +1141,28 @@ int main(int argc, char *argv[]) {
 			pcm_period_time = atoi(optarg);
 			break;
 
+		case '8' /* --volume */ : {
+
+			static const nv_entry_t values[] = {
+				{ "auto", .v.ui = VOL_TYPE_AUTO },
+				{ "mixer", .v.ui = VOL_TYPE_MIXER },
+				{ "software", .v.ui = VOL_TYPE_SOFTWARE },
+				{ "none", .v.ui = VOL_TYPE_NONE },
+				{ 0 },
+			};
+
+			const nv_entry_t *entry;
+			if ((entry = nv_find(values, optarg)) == NULL) {
+				error("Invalid volume control type {%s}: %s",
+						nv_join_names(values), optarg);
+				return EXIT_FAILURE;
+			}
+
+			volume_type_str = optarg;
+			volume_type = entry->v.ui;
+			break;
+		}
+
 		case 'M' /* --mixer-device=NAME */ :
 			mixer_device = optarg;
 			break;
@@ -1097,15 +1189,20 @@ int main(int argc, char *argv[]) {
 			return EXIT_FAILURE;
 		}
 
+	if ((main_loop_quit_event_fd = eventfd(0, EFD_CLOEXEC)) == -1) {
+		error("Couldn't create quit event: %s", strerror(errno));
+		return EXIT_FAILURE;
+	}
+
 	DBusError err = DBUS_ERROR_INIT;
-	if (!bluealsa_dbus_connection_ctx_init(&dbus_ctx, dbus_ba_service, &err)) {
+	if (!ba_dbus_connection_ctx_init(&dbus_ctx, dbus_ba_service, &err)) {
 		error("Couldn't initialize D-Bus context: %s", err.message);
 		return EXIT_FAILURE;
 	}
 
 	if (list_bt_devices || list_bt_pcms) {
 
-		if (!bluealsa_dbus_get_pcms(&dbus_ctx, &ba_pcms, &ba_pcms_count, &err)) {
+		if (!ba_dbus_pcm_get_all(&dbus_ctx, &ba_pcms, &ba_pcms_count, &err)) {
 			warn("Couldn't get BlueALSA PCM list: %s", err.message);
 			return EXIT_FAILURE;
 		}
@@ -1126,6 +1223,9 @@ int main(int argc, char *argv[]) {
 		return EXIT_FAILURE;
 	}
 
+	if (volume_type == VOL_TYPE_NONE || volume_type == VOL_TYPE_SOFTWARE)
+		mixer_device = NULL;
+
 	if (verbose >= 1) {
 
 		char *ba_str = malloc(19 * ba_addrs_count + 1);
@@ -1135,31 +1235,42 @@ int main(int argc, char *argv[]) {
 		for (i = 0; i < ba_addrs_count; i++, tmp += 19)
 			ba2str(&ba_addrs[i], stpcpy(tmp, ", "));
 
+		const char *mixer_device_str = "(not used)";
+		char mixer_element_str[128] = "(not used)";
+		if (mixer_device != NULL) {
+			mixer_device_str = mixer_device;
+			snprintf(mixer_element_str, sizeof(mixer_element_str), "'%s',%u",
+					mixer_elem_name, mixer_elem_index);
+		}
+
 		info("Selected configuration:\n"
 				"  BlueALSA service: %s\n"
 				"  ALSA PCM device: %s\n"
 				"  ALSA PCM buffer time: %u us\n"
 				"  ALSA PCM period time: %u us\n"
+				"  Volume control type: %s\n"
 				"  ALSA mixer device: %s\n"
-				"  ALSA mixer element: '%s',%u\n"
+				"  ALSA mixer element: %s\n"
 				"  Bluetooth device(s): %s\n"
 				"  Profile: %s",
 				dbus_ba_service,
 				pcm_device, pcm_buffer_time, pcm_period_time,
-				mixer_device, mixer_elem_name, mixer_elem_index,
+				volume_type_str,
+				mixer_device_str,
+				mixer_element_str,
 				ba_addr_any ? "ANY" : &ba_str[2],
 				ba_profile_a2dp ? "A2DP" : "SCO");
 
 		free(ba_str);
 	}
 
-	bluealsa_dbus_connection_signal_match_add(&dbus_ctx,
+	ba_dbus_connection_signal_match_add(&dbus_ctx,
 			dbus_ba_service, NULL, DBUS_INTERFACE_OBJECT_MANAGER, "InterfacesAdded",
 			"path_namespace='/org/bluealsa'");
-	bluealsa_dbus_connection_signal_match_add(&dbus_ctx,
+	ba_dbus_connection_signal_match_add(&dbus_ctx,
 			dbus_ba_service, NULL, DBUS_INTERFACE_OBJECT_MANAGER, "InterfacesRemoved",
 			"path_namespace='/org/bluealsa'");
-	bluealsa_dbus_connection_signal_match_add(&dbus_ctx,
+	ba_dbus_connection_signal_match_add(&dbus_ctx,
 			dbus_ba_service, NULL, DBUS_INTERFACE_PROPERTIES, "PropertiesChanged",
 			"arg0='"BLUEALSA_INTERFACE_PCM"'");
 
@@ -1168,41 +1279,49 @@ int main(int argc, char *argv[]) {
 		return EXIT_FAILURE;
 	}
 
-	if (!bluealsa_dbus_get_pcms(&dbus_ctx, &ba_pcms, &ba_pcms_count, &err))
+	if (!ba_dbus_pcm_get_all(&dbus_ctx, &ba_pcms, &ba_pcms_count, &err))
 		warn("Couldn't get BlueALSA PCM list: %s", err.message);
 
 	for (size_t i = 0; i < ba_pcms_count; i++)
 		supervise_io_worker(&ba_pcms[i]);
 
-	struct sigaction sigact = { .sa_handler = main_loop_stop };
+	struct sigaction sigact = {
+		.sa_handler = main_loop_stop,
+		.sa_flags = SA_RESETHAND };
+	/* Call to these handlers restores the default action, so on the
+	 * second call the program will be forcefully terminated. */
 	sigaction(SIGTERM, &sigact, NULL);
 	sigaction(SIGINT, &sigact, NULL);
 
 	debug("Starting main loop");
-	while (main_loop_on) {
+	for (;;) {
 
-		struct pollfd pfds[10];
-		nfds_t pfds_len = ARRAYSIZE(pfds);
+		struct pollfd fds[10] = {
+			{ main_loop_quit_event_fd, POLLIN, 0 } };
+		nfds_t nfds = ARRAYSIZE(fds) - 1;
 
-		if (!bluealsa_dbus_connection_poll_fds(&dbus_ctx, pfds, &pfds_len)) {
+		if (!ba_dbus_connection_poll_fds(&dbus_ctx, &fds[1], &nfds)) {
 			error("Couldn't get D-Bus connection file descriptors");
 			return EXIT_FAILURE;
 		}
 
-		if (poll(pfds, pfds_len, -1) == -1 &&
+		if (poll(fds, nfds + 1, -1) == -1 &&
 				errno == EINTR)
 			continue;
 
-		if (bluealsa_dbus_connection_poll_dispatch(&dbus_ctx, pfds, pfds_len))
+		if (fds[0].revents & POLLIN)
+			break;
+
+		if (ba_dbus_connection_poll_dispatch(&dbus_ctx, &fds[1], nfds))
 			while (dbus_connection_dispatch(dbus_ctx.conn) == DBUS_DISPATCH_DATA_REMAINS)
 				continue;
 
 	}
 
-	for (size_t i = 0; i < workers_count; i++)
-		io_worker_stop(i);
+	for (size_t i = workers_count; i > 0; i--)
+		io_worker_stop(i - 1);
 	free(workers);
 
-	bluealsa_dbus_connection_ctx_free(&dbus_ctx);
+	ba_dbus_connection_ctx_free(&dbus_ctx);
 	return EXIT_SUCCESS;
 }
